//
//  Little cms
//  Copyright (C) 1998-2000 Marti Maria
//
// THIS SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL MARTI MARIA BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "lcms.h"

// This module handles all formats supported by lcms

// typedef LPBYTE (* _cmsFIXFN)(WORD ToUnroll[], LPBYTE Buffer);

_cmsFIXFN cdecl _cmsIdentifyInputFormat(DWORD dwInput);
_cmsFIXFN cdecl _cmsIdentifyOutputFormat(DWORD dwOutput);

// ---------------------------------------------------------------------------


// This macro return words stored as big endian

#define CHANGE_ENDIAN(w)    (WORD) ((WORD) ((w)<<8)|((w)>>8))

// Unpacking routines.

static
LPBYTE Unroll4Bytes(WORD wIn[], LPBYTE accum)
{
       wIn[0] = RGB_8_TO_16(*accum); accum++;
       wIn[1] = RGB_8_TO_16(*accum); accum++;
       wIn[2] = RGB_8_TO_16(*accum); accum++;
       wIn[3] = RGB_8_TO_16(*accum); accum++;

       return accum;
}


// KCMY

static
LPBYTE Unroll4BytesSwap(WORD wIn[], LPBYTE accum)
{
       wIn[3] = RGB_8_TO_16(*accum); accum++;
       wIn[1] = RGB_8_TO_16(*accum); accum++;
       wIn[2] = RGB_8_TO_16(*accum); accum++;
       wIn[0] = RGB_8_TO_16(*accum); accum++;

       return accum;
}


static
LPBYTE Unroll4Words(WORD wIn[], LPBYTE accum)
{
       wIn[0] = *(LPWORD) accum; accum+= 2;
       wIn[1] = *(LPWORD) accum; accum+= 2;
       wIn[2] = *(LPWORD) accum; accum+= 2;
       wIn[3] = *(LPWORD) accum; accum+= 2;

       return accum;
}

static
LPBYTE Unroll4WordsSwap(WORD wIn[], LPBYTE accum)
{
       wIn[3] = *(LPWORD) accum; accum+= 2;
       wIn[1] = *(LPWORD) accum; accum+= 2;
       wIn[2] = *(LPWORD) accum; accum+= 2;
       wIn[0] = *(LPWORD) accum; accum+= 2;

       return accum;
}


static
LPBYTE Unroll4WordsBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[0] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[1] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[3] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;

       return accum;
}

static
LPBYTE Unroll4WordsSwapBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[3] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[1] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[0] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;

       return accum;
}

static
LPBYTE Unroll3Bytes(WORD wIn[], LPBYTE accum)
{

       wIn[0] = RGB_8_TO_16(*accum); accum++;     // R
       wIn[1] = RGB_8_TO_16(*accum); accum++;     // G
       wIn[2] = RGB_8_TO_16(*accum); accum++;     // B

       return accum;
}


// BRG

static
LPBYTE Unroll3BytesSwap(WORD wIn[], LPBYTE accum)
{

       wIn[2] = RGB_8_TO_16(*accum); accum++;     // B
       wIn[1] = RGB_8_TO_16(*accum); accum++;     // G
       wIn[0] = RGB_8_TO_16(*accum); accum++;     // R

       return accum;
}

static
LPBYTE Unroll3Words(WORD wIn[], LPBYTE accum)
{
       wIn[0] = *(LPWORD) accum; accum+= 2;  // C R
       wIn[1] = *(LPWORD) accum; accum+= 2;  // M G
       wIn[2] = *(LPWORD) accum; accum+= 2;  // Y B
       return accum;
}


static
LPBYTE Unroll3WordsSwap(WORD wIn[], LPBYTE accum)
{
       wIn[2] = *(LPWORD) accum; accum+= 2;  // C R
       wIn[1] = *(LPWORD) accum; accum+= 2;  // M G
       wIn[0] = *(LPWORD) accum; accum+= 2;  // Y B
       return accum;
}


static
LPBYTE Unroll3WordsBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[0] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[1] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       return accum;
}


static
LPBYTE Unroll3WordsSwapBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[1] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[0] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       return accum;
}




// Monochrome duplicates L into RGB for null-transforms

static
LPBYTE Unroll1Byte(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = RGB_8_TO_16(*accum); accum++;     // L
       return accum;
}


static
LPBYTE Unroll1Word(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = *(LPWORD) accum; accum+= 2;   // L
       return accum;
}


static
LPBYTE Unroll1WordBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       return accum;
}

// Monochrome + alpha. Alpha is lost

static
LPBYTE Unroll2Byte(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = RGB_8_TO_16(*accum); accum++;     // L
       wIn[3] = RGB_8_TO_16(*accum); accum++;                       // alpha
       return accum;
}


static
LPBYTE Unroll2Word(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = *(LPWORD) accum; accum+= 2;   // L
       wIn[3] = *(LPWORD) accum; accum += 2;                    // alpha

       return accum;
}


static
LPBYTE Unroll2WordBigEndian(WORD wIn[], LPBYTE accum)
{
       wIn[0] = wIn[1] = wIn[2] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;
       wIn[3] = CHANGE_ENDIAN(*(LPWORD) accum); accum+= 2;

       return accum;
}



// Packing routines

static
LPBYTE Pack6Bytes(WORD wOut[], LPBYTE output)
{
              *output++ = RGB_16_TO_8(wOut[0]);
              *output++ = RGB_16_TO_8(wOut[1]);
              *output++ = RGB_16_TO_8(wOut[2]);
              *output++ = RGB_16_TO_8(wOut[3]);
              *output++ = RGB_16_TO_8(wOut[4]);
              *output++ = RGB_16_TO_8(wOut[5]);


              return output;
}


static
LPBYTE Pack6Words(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[2];
       output+= 2;
       *(LPWORD) output = wOut[3];
       output+= 2;
       *(LPWORD) output = wOut[4];
       output+= 2;
       *(LPWORD) output = wOut[5];
       output+= 2;

       return output;
}

static
LPBYTE Pack4Bytes(WORD wOut[], LPBYTE output)
{
              *output++ = RGB_16_TO_8(wOut[0]);
              *output++ = RGB_16_TO_8(wOut[1]);
              *output++ = RGB_16_TO_8(wOut[2]);
              *output++ = RGB_16_TO_8(wOut[3]);

              return output;
}


// ABGR
static
LPBYTE Pack4BytesSwap(WORD wOut[], LPBYTE output)
{
              *output++ = RGB_16_TO_8(wOut[2]);
              *output++ = RGB_16_TO_8(wOut[1]);
              *output++ = RGB_16_TO_8(wOut[0]);
              *output++ = RGB_16_TO_8(wOut[3]);

              return output;
}


static
LPBYTE Pack4Words(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[2];
       output+= 2;
       *(LPWORD) output = wOut[3];
       output+= 2;

       return output;
}


static
LPBYTE Pack4WordsSwap(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[2];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[0];
       output+= 2;
       *(LPWORD) output = wOut[3];
       output+= 2;

       return output;
}

static
LPBYTE Pack3Bytes(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[0]);
       *output++ = RGB_16_TO_8(wOut[1]);
       *output++ = RGB_16_TO_8(wOut[2]);

       return output;
}

static
LPBYTE Pack3BytesSwap(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[2]);
       *output++ = RGB_16_TO_8(wOut[1]);
       *output++ = RGB_16_TO_8(wOut[0]);

       return output;
}


static
LPBYTE Pack3Words(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[2];
       output+= 2;

       return output;
}

static
LPBYTE Pack3WordsSwap(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[2];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[0];
       output+= 2;

       return output;
}

static
LPBYTE Pack3WordsBigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = CHANGE_ENDIAN(wOut[0]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[1]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[2]);
       output+= 2;

       return output;
}


static
LPBYTE Pack3WordsSwapBigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = CHANGE_ENDIAN(wOut[2]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[1]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[0]);
       output+= 2;

       return output;
}


static
LPBYTE Pack3BytesAndSkip1(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[0]);
       *output++ = RGB_16_TO_8(wOut[1]);
       *output++ = RGB_16_TO_8(wOut[2]);
       output++;

       return output;
}

static
LPBYTE Pack3BytesAndSkip1Swap(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[2]);
       *output++ = RGB_16_TO_8(wOut[1]);
       *output++ = RGB_16_TO_8(wOut[0]);
       output++;

       return output;
}



static
LPBYTE Pack3WordsAndSkip1(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[2];
       output+= 2;
       output+= 2;

       return output;
}

static
LPBYTE Pack3WordsAndSkip1Swap(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[2];
       output+= 2;
       *(LPWORD) output = wOut[1];
       output+= 2;
       *(LPWORD) output = wOut[0];
       output+= 2;
       output+= 2;

       return output;
}


static
LPBYTE Pack3WordsAndSkip1BigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = CHANGE_ENDIAN(wOut[0]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[1]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[2]);
       output+= 2;
       output+= 2;

       return output;
}


static
LPBYTE Pack3WordsAndSkip1SwapBigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = CHANGE_ENDIAN(wOut[2]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[1]);
       output+= 2;
       *(LPWORD) output = CHANGE_ENDIAN(wOut[0]);
       output+= 2;
       output+= 2;

       return output;
}



static
LPBYTE Pack1Byte(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[0]);
       return output;
}


static
LPBYTE Pack1ByteAndSkip1(WORD wOut[], LPBYTE output)
{
       *output++ = RGB_16_TO_8(wOut[0]);
       output++;
       return output;
}

static
LPBYTE Pack1Word(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;

       return output;
}

static
LPBYTE Pack1WordBigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 2;

       return output;
}


static
LPBYTE Pack1WordAndSkip1(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = wOut[0];
       output+= 4;

       return output;
}


static
LPBYTE Pack1WordAndSkip1BigEndian(WORD wOut[], LPBYTE output)
{
       *(LPWORD) output = CHANGE_ENDIAN(wOut[0]);
       output+= 4;

       return output;
}


//  choose routine from Input identifier

_cmsFIXFN _cmsIdentifyInputFormat(DWORD dwInput)
{
              _cmsFIXFN FromInput = NULL;

              switch (T_BYTES(dwInput)) {

              case 1: // 1 byte per channel

                     switch (T_CHANNELS(dwInput) + T_EXTRA(dwInput))
                     {
                     case 1: FromInput = Unroll1Byte;
                             break;

                     case 2: FromInput = Unroll2Byte;
                             break;

                     case 3: if (T_DOSWAP(dwInput))
                                   FromInput = Unroll3BytesSwap;
                             else
                                   FromInput = Unroll3Bytes;
                             break;
                     case 4:
                             if (T_DOSWAP(dwInput))
                                   FromInput = Unroll4BytesSwap;
                             else
                                   FromInput = Unroll4Bytes;
                             break;
                     default:;
                     }
                     break;


              case 2: // 1 word per channel

                     switch (T_CHANNELS(dwInput) + T_EXTRA(dwInput))
                     {
                     case 1: if (T_ENDIAN16(dwInput))
                                   FromInput = Unroll1WordBigEndian;
                             else
                                   FromInput = Unroll1Word;
                             break;

                     case 2: if (T_ENDIAN16(dwInput))
                                   FromInput = Unroll2WordBigEndian;
                             else
                                   FromInput = Unroll2Word;
                             break;

                     case 3: if (T_DOSWAP(dwInput)) {
                                   if (T_ENDIAN16(dwInput))
                                          FromInput = Unroll3WordsSwapBigEndian;
                                   else
                                          FromInput = Unroll3WordsSwap;
                             }
                             else {
                                   if (T_ENDIAN16(dwInput))
                                          FromInput = Unroll3WordsBigEndian;
                                   else
                                          FromInput = Unroll3Words;
                             }
                             break;

                     case 4: if (T_DOSWAP(dwInput)) {
                                   if (T_ENDIAN16(dwInput))
                                          FromInput = Unroll4WordsSwapBigEndian;
                                   else
                                          FromInput = Unroll4WordsSwap;
                             }
                             else {
                                   if (T_ENDIAN16(dwInput))
                                          FromInput = Unroll4WordsBigEndian;
                                   else
                                          FromInput = Unroll4Words;
                             }
                             break;
                     }
                     break;

              default:;
              }

       if (!FromInput)
              cmsSignalError(LCMS_ERRC_ABORTED, "Unknown input format");

       return FromInput;

}

//  choose routine from Input identifier

_cmsFIXFN _cmsIdentifyOutputFormat(DWORD dwOutput)
{
              _cmsFIXFN ToOutput = NULL;

              switch (T_BYTES(dwOutput)) {

              case 1:
                     switch (T_CHANNELS(dwOutput))
                     {
                     case 1:
                            ToOutput = Pack1Byte;
                            if (T_EXTRA(dwOutput) == 1)
                                   ToOutput = Pack1ByteAndSkip1;
                            break;

                     case 3:
                         switch (T_EXTRA(dwOutput)) {

                         case 0: if (T_DOSWAP(dwOutput))
                                   ToOutput = Pack3BytesSwap;
                             else
                                   ToOutput = Pack3Bytes;
                             break;

                         case 1: if (T_DOSWAP(dwOutput))
                                   ToOutput = Pack3BytesAndSkip1Swap;
                             else
                                   ToOutput = Pack3BytesAndSkip1;
                             break;

                         default:;
                         }
                         break;

                     case 4: if (T_EXTRA(dwOutput) == 0)
                             {
                             if (T_DOSWAP(dwOutput))
                                   ToOutput = Pack4BytesSwap;
                             else
                                   ToOutput = Pack4Bytes;
                             }
                             break;

                     // Hexachrome separations.
                     case 6: if (T_EXTRA(dwOutput) == 0 && !T_DOSWAP(dwOutput))
                                   ToOutput = Pack6Bytes;
                             break;

                     default:;
                     }
                     break;


              case 2:

                     switch (T_CHANNELS(dwOutput)) {

                     case 1:
                            if (T_ENDIAN16(dwOutput))

                                   ToOutput = Pack1WordBigEndian;
                            else
                                   ToOutput = Pack1Word;

                            if (T_EXTRA(dwOutput) == 1) {

                               if (T_ENDIAN16(dwOutput))

                                   ToOutput = Pack1WordAndSkip1BigEndian;
                               else
                                   ToOutput = Pack1WordAndSkip1;
                            }
                            break;

                     case 3:

                         switch (T_EXTRA(dwOutput)) {

                         case 0:
                               if (T_DOSWAP(dwOutput)) {

                                   if (T_ENDIAN16(dwOutput))

                                          ToOutput = Pack3WordsSwapBigEndian;
                                   else
                                          ToOutput = Pack3WordsSwap;
                               }
                               else {
                                   if (T_ENDIAN16(dwOutput))

                                      ToOutput = Pack3WordsBigEndian;
                                   else
                                      ToOutput = Pack3Words;
                                   }
                             break;

                         case 1: if (T_DOSWAP(dwOutput)) {

                                   if (T_ENDIAN16(dwOutput))

                                          ToOutput = Pack3WordsAndSkip1SwapBigEndian;
                                   else
                                          ToOutput = Pack3WordsAndSkip1Swap;
                             }
                             else  {
                                   if (T_ENDIAN16(dwOutput))
                                          ToOutput = Pack3WordsAndSkip1BigEndian;
                                   else
                                          ToOutput = Pack3WordsAndSkip1;
                                   }
                         default:;
                         }
                         break;

                     case 4: if (T_EXTRA(dwOutput) == 0)
                            {
                            if (T_DOSWAP(dwOutput))
                                   ToOutput = Pack4WordsSwap;
                            else
                                   ToOutput = Pack4Words;
                            }
                            break;

                     case 6: if (T_EXTRA(dwOutput) == 0
                                   && !T_DOSWAP(dwOutput)
                                   && !T_ENDIAN16(dwOutput))
                                          ToOutput = Pack6Words;
                             break;

                     default:;
                     }
                     break;

              default:;
              }

              if (!ToOutput)
                     cmsSignalError(LCMS_ERRC_ABORTED, "Unknown output format");

              return ToOutput;
}

